/*
 *    Copyright (C) 2006 Richard Kotal
 *
 *    This library is free software; you can redistribute it and/or modify it 
 *    under the terms of the GNU Library General Public License as published 
 *    by the Free Software Foundation; either version 2 of the License, or 
 *    (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this library; if not, write to the
 *    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *    Boston, MA 02111-1307, USA.
 */  
  
static int
ns_mcrypt_TclInterpInit (Tcl_Interp * interp, void *ctx)
{
  Tcl_CreateObjCommand (interp, TCLCMD, ns_mcrypt_TclCmdRegister, ctx, NULL);
  Tcl_RegisterObjType(&tclMcryptType);

  return NS_OK;
}


static int
ns_mcrypt_TclCmdRegister (ClientData arg, Tcl_Interp * interp, int objc,
			  Tcl_Obj * CONST objv[])
{
  static CONST2 char *commands[] =
    { "algorithms", "modes", "capability", "new", "destroy", "initvector" , "init", "flush", "crypt", "decrypt", "detach", "attach", NULL };
  enum
  { AlgoIdx, ModesIdx, CapaIdx, NewIdx, DestroyIdx, InitVectorIdx, InitIdx, FlushIdx, CryptIdx, DecryptIdx , PutShareIdx, GetShareIdx}
  cmd;

  if (objc < 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "command ? option arg ... ?");
      return TCL_ERROR;
    }
  if (Tcl_GetIndexFromObj
      (interp, objv[1], commands, "command", 0, (int *) &cmd) != TCL_OK)
    return TCL_ERROR;
  switch (cmd)
    {
    case AlgoIdx:
      {
	char *dir = NULL;
	if (objc > 3)
	  {
	    Tcl_WrongNumArgs (interp, 2, objv, "?directory_of_algorithms?");
	    return TCL_ERROR;
	  }
	if (objc == 3)
	  dir = Tcl_GetString (objv[2]);
	ns_mcrypt_AlgoModes (interp, STR_ALGO, dir);
	break;
      }				/* AlgoIdx */
    case ModesIdx:
      {
	char *dir = NULL;
	if (objc > 3)
	  {
	    Tcl_WrongNumArgs (interp, 2, objv, "?directory_of_modes?");
	    return TCL_ERROR;
	  }
	if (objc == 3)
	  dir = Tcl_GetString (objv[2]);
	ns_mcrypt_AlgoModes (interp, STR_MODES, dir);
	break;
      }				/* ModesIdx */
    case CapaIdx:
      {
	char *algoname = NULL;
	char *type = STR_ALGO;
	char *dir = NULL;
	if (objc < 4 || objc > 5)
	  {
	    Tcl_WrongNumArgs (interp, 2, objv,
			      "-algo|-mode algorithm_or_mode_name ?directory_of_modes_or_algo?");
	    return TCL_ERROR;
	  }
	if (STREQ (Tcl_GetString (objv[2]), "-mode"))
	  type = STR_MODES;
	algoname = Tcl_GetString (objv[3]);
	if (objc == 5)
	  dir = Tcl_GetString (objv[4]);
	return ns_mcrypt_Capability (interp, algoname, dir, type);
	break;
      }				// CapaIdx //
    case NewIdx:
      {
	char *algoname = NULL;
	char *modename = NULL;
	char *algodir = NULL;
	char *modedir = NULL;
	if (objc < 4 || objc > 6)
	  {
	    Tcl_WrongNumArgs (interp, 2, objv, "algorithm_name mode_name ?algo_directory? ?mode_directory?");
	    return TCL_ERROR;
	  }
	algoname = Tcl_GetString (objv[2]);
	modename = Tcl_GetString (objv[3]);
	if (objc > 4) algodir = Tcl_GetString (objv[4]);
	if (objc > 5) modedir = Tcl_GetString (objv[5]);
	return ns_mcrypt_NewObj (interp, algoname,modename,algodir,modedir);
	break;
      }				// NewIdx //
    case DestroyIdx:
      {
	if (objc != 3)
	  {
	    Tcl_WrongNumArgs (interp, 2, objv, "mcryptObjId");
	    return TCL_ERROR;
	  }
	return ns_mcrypt_DestroyObj (objv[2]);
	break;
      }				// DestroyIdx //

    case InitVectorIdx:
      {
	static CONST2 char *opts[] = { "set", "get", NULL };
	enum { IVSetIdx, IVGetIdx } opt;
	if (objc < 4)
	  {
	    Tcl_WrongNumArgs (interp, 2, objv, "command ? option arg ... ?");
	    return TCL_ERROR;
	  }
	if (Tcl_GetIndexFromObj (interp, objv[2], opts, "command for initvector", 0, (int *) &opt) != TCL_OK)
	  return TCL_ERROR;
	switch (opt) {
	    case IVSetIdx: {
	        int bin = -1;
		char *IV = NULL;
		int size = 0;
	        if (objc < 5 || objc > 6)
		{
		  Tcl_WrongNumArgs (interp, 3, objv, "-raw|-base64|-hex|-random mcryptObjId ?initvector_string?");
		  return TCL_ERROR;
		}
	        if (STREQ (Tcl_GetString (objv[3]), "-raw")) bin = 1;
	        else if (STREQ (Tcl_GetString (objv[3]), "-base64")) bin = 2;
	        else if (STREQ (Tcl_GetString (objv[3]), "-hex")) bin = 3;
	        else if (STREQ (Tcl_GetString (objv[3]), "-random")) bin = 4;
		if (objc > 5)  IV = Tcl_GetByteArrayFromObj (objv[5], &size);
		return ns_mcrypt_SetIV(interp,objv[4],bin,IV,size);
	        break;
	    } // IVSetIdx //
	    case IVGetIdx: {
	        int bin = -1;
	        if (objc != 5 )
		{
		  Tcl_WrongNumArgs (interp, 3, objv, "-raw|-base64|-hex mcryptObjId ");
		  return TCL_ERROR;
		}
	        if (STREQ (Tcl_GetString (objv[3]), "-raw")) bin = 1;
	        else if (STREQ (Tcl_GetString (objv[3]), "-base64")) bin = 2;
	        else if (STREQ (Tcl_GetString (objv[3]), "-hex")) bin = 3;
		return ns_mcrypt_GetIV(interp,objv[4],bin);
	        break;
	    } // IVGetIdx //
	} // opt - InitVectorIdx //
	break;
      }				// InitVectorIdx //
    case InitIdx:
      {
	char *key = NULL;
	int keysize = 0;
	if (objc != 4)
	  {
	    Tcl_WrongNumArgs (interp, 2, objv, " mcryptObjId keyword");
	    return TCL_ERROR;
	  }
	key = Tcl_GetByteArrayFromObj (objv[3], &keysize);
	return ns_mcrypt_InitObj (interp, objv[2], key,keysize);
	break;
      }				// InitIdx //
    case FlushIdx:
      {
	if (objc != 3)
	  {
	    Tcl_WrongNumArgs (interp, 2, objv, "mcryptObjId");
	    return TCL_ERROR;
	  }
	return ns_mcrypt_FlushObj (interp, objv[2]);
	break;
      }				// FlushIdx //
    case CryptIdx:
    case DecryptIdx:
      {
	char *text = NULL;
	int textsize = 0;
	int in_bin = -1;
	int out_bin = -1;
	if (objc != 8)
	  {
	    Tcl_WrongNumArgs (interp, 2, objv, " -in raw|hex|base64 -out str|raw|hex|base64 mcryptObjId plaintext|ciphertext");
	    return TCL_ERROR;
	  }
        if (!(STREQ(Tcl_GetString (objv[2]), "-in") && STREQ(Tcl_GetString (objv[4]), "-out")))
	  {
	    Tcl_WrongNumArgs (interp, 2, objv, " -in raw|hex|base64 -out str|raw|hex|base64 mcryptObjId plaintext|ciphertext");
	    return TCL_ERROR;
	  }
	if (STREQ (Tcl_GetString (objv[3]), "raw")) in_bin = 1;
	else if (STREQ (Tcl_GetString (objv[3]), "base64")) in_bin = 2;
	else if (STREQ (Tcl_GetString (objv[3]), "hex")) in_bin = 3;

	if (STREQ (Tcl_GetString (objv[5]), "raw")) out_bin = 1;
	else if (STREQ (Tcl_GetString (objv[5]), "base64")) out_bin = 2;
	else if (STREQ (Tcl_GetString (objv[5]), "hex")) out_bin = 3;
	else if (STREQ (Tcl_GetString (objv[5]), "str")) out_bin = 4;
	  
	text = Tcl_GetByteArrayFromObj (objv[7], &textsize);
	return ns_mcrypt_CryptObj (interp, objv[6],(cmd==CryptIdx)?CRYPT:DECRYPT, text,textsize,in_bin,out_bin);
	break;
      }				// CryptIdx, DecryptIdx //
    case PutShareIdx:
     {
      if (objc != 3)
	{
	  Tcl_WrongNumArgs (interp, 2, objv, "mcryptObjId");
	  return TCL_ERROR;
	}
      return ns_mcrypt_PutShareCryptObj (interp, objv[2]);
      break;
     }			// PutShareIdx //
    case GetShareIdx:
     {
      if (objc != 3)
	{
	  Tcl_WrongNumArgs (interp, 2, objv, "mcryptDetachId");
	  return TCL_ERROR;
	}
      return ns_mcrypt_GetShareCryptObj (interp, objv[2]);
      break;
     }			// GetShareIdx //


    }				/* cmd */
  return TCL_OK;
}
